import logging
import typing

from atomic_agents.agents.base_agent import (
    BaseAgent,
    BaseAgentConfig,
    BaseIOSchema,
)
from cornsnake import util_time, util_wait
from rich.console import Console

from gpt_multi_atomic_agents.graphql_dto import GraphQLAgentOutputSchema

from .functions_dto import FunctionAgentOutputSchema
from .util_output import print_warning

from .blackboard_serde import load_blackboard_from_file

from .prompts_router import AgentExecutionPlanSchema

from . import main_router

from . import util_ai
from .agent_definition import (
    AgentDefinitionBase,
    FunctionAgentDefinition,
)
from .blackboard import (
    Blackboard,
    FunctionCallBlackboard,
    GraphQLBlackboard,
    Message,
    MessageRole,
)
from .blackboard_accessor import (
    BlackboardAccessor,
    FunctionCallBlackboardAccessor,
    GraphQLBlackboardAccessor,
)
from .config import Config
from .repl_commands import CommandAction, check_user_prompt, print_help
from . import util_print_agent

console = Console()

logger = logging.getLogger(__file__)


def _create_agent(agent_definition: AgentDefinitionBase, _config: Config) -> BaseAgent:
    client, model, max_tokens = util_ai.create_client(_config=_config)
    system_prompt_builder = agent_definition.get_system_prompt_builder(_config=_config)

    agent = BaseAgent(
        config=BaseAgentConfig(
            client=client,
            model=model,
            system_prompt_generator=system_prompt_builder.build_system_prompt(),
            input_schema=agent_definition.input_schema,
            output_schema=agent_definition.output_schema,
            max_tokens=max_tokens,
        )
    )
    return agent


def _check_blackboard(
    blackboard: Blackboard, agent_definitions: list[AgentDefinitionBase]
) -> None:
    is_function_based = isinstance(agent_definitions[0], FunctionAgentDefinition)
    if blackboard:
        if is_function_based:
            if not (typing.cast(FunctionCallBlackboard, blackboard)):
                raise RuntimeError("Expected blackboard to be a FunctionCallBlackboard")
        elif not (typing.cast(GraphQLBlackboard, blackboard)):
            raise RuntimeError("Expected blackboard to be a GraphQLBlackboard")


def _create_blackboard_accessor(
    agent_definitions: list[AgentDefinitionBase],
) -> BlackboardAccessor:
    if not agent_definitions:
        raise RuntimeError("Expected at least 1 Agent Definition")
    is_function_based = isinstance(agent_definitions[0], FunctionAgentDefinition)
    blackboard = (
        FunctionCallBlackboardAccessor(_blackboard=FunctionCallBlackboard())
        if is_function_based
        else GraphQLBlackboardAccessor(_blackboard=GraphQLBlackboard())
    )
    return blackboard


def _create_blackboard_accessor_from_blackboard(
    blackboard: Blackboard,
) -> BlackboardAccessor:
    if isinstance(blackboard, FunctionCallBlackboard):
        return FunctionCallBlackboardAccessor(_blackboard=blackboard)
    elif isinstance(blackboard, GraphQLBlackboard):
        return GraphQLBlackboardAccessor(_blackboard=blackboard)

    raise RuntimeError("Not a recognised kind of Blackboard")


def generate(
    agent_definitions: list[AgentDefinitionBase],
    chat_agent_description: str,
    _config: Config,
    user_prompt: str,
    blackboard: (
        BlackboardAccessor | None
    ) = None,  # If used as a web service, then would also accept previous state + new data (which the user has updated either by executing its implementation of Function Calls OR by updating via GraphQL mutations).
    execution_plan: AgentExecutionPlanSchema | None = None,
) -> BlackboardAccessor:
    previous_blackboard = blackboard._blackboard if blackboard else None
    new_blackboard = generate_with_blackboard(
        agent_definitions=agent_definitions,
        chat_agent_description=chat_agent_description,
        _config=_config,
        user_prompt=user_prompt,
        blackboard=previous_blackboard,
        execution_plan=execution_plan,
    )
    return _create_blackboard_accessor_from_blackboard(blackboard=new_blackboard)


def _create_blackboard(
    agent_definitions: list[AgentDefinitionBase],
) -> Blackboard:
    if not agent_definitions:
        raise RuntimeError("Expected at least 1 Agent Definition")
    is_function_based = isinstance(agent_definitions[0], FunctionAgentDefinition)
    blackboard = FunctionCallBlackboard() if is_function_based else GraphQLBlackboard()
    return blackboard


def _fix_agent_name(
    response: BaseIOSchema, agent_definition: AgentDefinitionBase
) -> None:
    """Sometimes generate() hallucincates a new Agent Name (but planning does not!). We can easily post-fix it here."""
    if isinstance(response, FunctionAgentOutputSchema):
        did_warn = False
        for fun in response.generated_function_calls:
            if fun.agent_name != agent_definition.agent_name:
                if not did_warn:
                    print_warning(
                        message=f"LLM reponse has incorrect agent name '{fun.agent_name}' - fixing it to be '{agent_definition.agent_name}'"
                    )
                    did_warn = True
                fun.agent_name = agent_definition.agent_name
    elif isinstance(response, GraphQLAgentOutputSchema):
        pass  # no agent name in response
    else:
        raise RuntimeError("Not a recognised agent response")


PROCEED_PROMPT = "proceed"


def is_user_prompt_proceed(user_prompt: str) -> bool:
    user_prompt = user_prompt.strip()
    return not user_prompt or user_prompt.strip() == PROCEED_PROMPT


def _has_new_user_prompt(user_prompt: str) -> bool:
    # In theory, we could auto detect if the new user prompt really needs a new plan, but this seems tricky.
    return not is_user_prompt_proceed(user_prompt=user_prompt)


def generate_with_blackboard(
    agent_definitions: list[AgentDefinitionBase],
    chat_agent_description: str,
    _config: Config,
    user_prompt: str,
    blackboard: (
        Blackboard | None
    ) = None,  # If used as a web service, then would also accept previous state + new data (which the user has updated either by executing its implementation of Function Calls OR by updating via GraphQL mutations).
    execution_plan: AgentExecutionPlanSchema | None = None,
) -> Blackboard:
    """
    Use the provided agents to fulfill the user's prompt.
    - if an execution plan is provided, that is used to decide which agents to execute.
       - else the router is used to generate an execution plan
    - if a user prompt is provided, then a new execution plan is generated (since the user may need different agents).
    """

    start = util_time.start_timer()

    if blackboard:
        _check_blackboard(blackboard=blackboard, agent_definitions=agent_definitions)
    else:
        blackboard = _create_blackboard(agent_definitions)

    blackboard.add_previous_message(Message(role=MessageRole.user, message=user_prompt))

    with console.status("[bold green]Processing...") as _status:
        try:
            if (
                not execution_plan or _has_new_user_prompt(user_prompt=user_prompt)
            ):  # A new user prompt means we likely need a new plan, for example if different agents are needed.
                if execution_plan:
                    print_warning(
                        "Generating a new plan: Generate received a user prompt, so discarding the current generation plan (to optimize, you can send a plan with an empty user prompt)"
                    )
                else:
                    print("Generating a new plan")
                execution_plan = main_router.generate_plan(
                    agent_definitions=agent_definitions,
                    chat_agent_description=chat_agent_description,
                    _config=_config,
                    user_prompt=user_prompt,
                    previous_plan=None,
                    messages=blackboard.internal_previous_messages,
                )
                blackboard.add_message(
                    Message(
                        role=MessageRole.assistant, message=execution_plan.chat_message
                    )
                )
                util_print_agent.print_assistant_message(execution_plan.chat_message)
                util_wait.wait_seconds(_config.delay_between_calls_in_seconds)

            # Loop thru all the recommended agents, sending each one a rewritten version of the user prompt
            for i, recommended_agent in enumerate(execution_plan.recommended_agents):
                try:
                    if recommended_agent.agent_name == "chat":
                        # TODO: add option to redirect to some Chat agent
                        continue

                    console.log(
                        f":robot: Executing agent {recommended_agent.agent_name}..."
                    )
                    util_print_agent.print_agent(
                        recommended_agent, _config=_config, prefix="EXECUTING: "
                    )
                    matching_agent_definitions = list(
                        filter(
                            lambda a: a.agent_name == recommended_agent.agent_name,
                            agent_definitions,
                        )
                    )
                    if not matching_agent_definitions:
                        raise RuntimeError(
                            f"Could not match recommended agent {recommended_agent.agent_name}"
                        )
                    if len(matching_agent_definitions) > 1:
                        print_warning(
                            f"Matched more than one agent to {recommended_agent.agent_name}"
                        )
                    agent_definition = matching_agent_definitions[0]
                    agent = _create_agent(agent_definition, _config=_config)

                    response = agent.run(
                        agent_definition.build_input(
                            recommended_agent.rewritten_user_prompt,
                            blackboard=blackboard,
                            config=_config,
                            agent_parameters=recommended_agent.agent_parameters,
                        )
                    )
                    _fix_agent_name(response, agent_definition)
                    util_print_agent.print_assistant_output(response, agent_definition)

                    agent_definition.update_blackboard(
                        response=response, blackboard=blackboard
                    )
                    is_last = i == len(execution_plan.recommended_agents) - 1
                    if not is_last:
                        util_wait.wait_seconds(_config.delay_between_calls_in_seconds)
                except Exception as e:
                    logger.exception(e)
        except Exception as e:
            logger.exception(e)

        console.log(":robot: (done)")
        time_taken = util_time.end_timer(start=start)
        console.log(f"  time taken: {util_time.describe_elapsed_seconds(time_taken)}")
    return blackboard


def run_chat_loop(
    agent_definitions: list[AgentDefinitionBase],
    chat_agent_description: str,
    _config: Config,
    given_user_prompt: str | None = None,
    blackboard: (
        BlackboardAccessor | None
    ) = None,  # If used as a web service, then would also accept previous state + new data (which the user has updated either by executing its implementation of Function Calls OR by updating via GraphQL mutations).
) -> BlackboardAccessor:
    """
    Use the provided agents to fulfill the user's prompt.
    - if an execution plan is provided, that is used to decide which agents to execute.
       - else the router is used to generate an execution plan
    - if no given user prompt is provided, then the user is prompted via keyboard input in a REPL loop.
    """

    if not blackboard:
        blackboard = _create_blackboard_accessor(agent_definitions)

    print_help()

    initial_assistant_message = "How can I help you?"
    util_print_agent.print_assistant_message_only(
        chat_message=initial_assistant_message
    )

    blackboard._blackboard.add_message(
        Message(role=MessageRole.assistant, message=initial_assistant_message)
    )

    # for more emojis - see "poetry run python -m rich.emoji"
    if given_user_prompt:
        util_print_agent.print_user_prompt(given_user_prompt)

    while True:
        user_prompt = (
            given_user_prompt
            if given_user_prompt
            else console.input(":sunglasses: You: ")
        )
        action = check_user_prompt(
            user_prompt=user_prompt, blackboard=blackboard._blackboard, config=_config
        )
        match action:
            case CommandAction.quit:
                util_print_agent.print_assistant_message_only(chat_message="Good bye!")
                break
            case CommandAction.handled_already:
                continue
            case CommandAction.load_blackboard:
                new_blackboard = load_blackboard_from_file(
                    config=_config, existing_blackboard=blackboard._blackboard
                )
                if new_blackboard:
                    blackboard = _create_blackboard_accessor_from_blackboard(
                        blackboard=new_blackboard
                    )
                    console.print("(Blackboard loaded)")
                continue
            case CommandAction.no_action:
                pass
            case _:
                raise RuntimeError(f"Not a recognised CommandAction: {action}")

        blackboard = generate(
            agent_definitions=agent_definitions,
            chat_agent_description=chat_agent_description,
            _config=_config,
            user_prompt=user_prompt,
            blackboard=blackboard,
        )

        if given_user_prompt:
            break
    # To support a stateless web service, we return the whole blackboard, and accept it as optional input
    return blackboard


# to debug - see agent.system_prompt_generator.generate_prompt()



from abc import ABC, abstractmethod
from enum import StrEnum, auto
from typing import Text
from rich.console import Console

from .blackboard import Blackboard
from .blackboard_serde import list_blackboard_files, save_blackboard_to_file
from .config import Config
from .util_print_agent import print_assistant_message_only

console = Console()


class CommandAction(StrEnum):
    no_action = auto()
    handled_already = auto()
    load_blackboard = auto()
    quit = auto()


class ReplCommandBase(ABC):
    @abstractmethod
    def get_name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_description(self) -> str:
        raise NotImplementedError

    def get_aliases(self) -> list[str]:
        return []

    @abstractmethod
    def do(self, blackboard: Blackboard, config: Config) -> CommandAction:
        raise NotImplementedError


class ClearReplCommand(ReplCommandBase):
    def get_name(self) -> str:
        return "clear"

    def get_aliases(self) -> list[str]:
        return ["reset"]

    def get_description(self) -> str:
        return "Clear the blackboard, starting over."

    def do(self, blackboard: Blackboard, config: Config) -> CommandAction:
        blackboard.reset_all()
        console.print("(Blackboard has been reset)")
        return CommandAction.handled_already


class DumpReplCommand(ReplCommandBase):
    def get_name(self) -> str:
        return "dump"

    def get_aliases(self) -> list[str]:
        return ["show"]

    def get_description(self) -> str:
        return "Dump the current blackboard state to the console"

    def do(self, blackboard: Blackboard, config: Config) -> CommandAction:
        console.print(blackboard)
        return CommandAction.handled_already


class HelpReplCommand(ReplCommandBase):
    def get_name(self) -> str:
        return "help"

    def get_description(self) -> str:
        return "Display help text"

    def do(self, blackboard: Blackboard, config: Config) -> CommandAction:
        print_help()
        return CommandAction.handled_already


class ListReplCommand(ReplCommandBase):
    def get_name(self) -> str:
        return "list"

    def get_description(self) -> str:
        return "List the local data files from previous blackboards"

    def do(self, blackboard: Blackboard, config: Config) -> CommandAction:
        list_blackboard_files(blackboard=blackboard, config=config)
        return CommandAction.handled_already


class LoadReplCommand(ReplCommandBase):
    def get_name(self) -> str:
        return "load"

    def get_description(self) -> str:
        return "Load a blackboard from the local data store"

    def do(self, blackboard: Blackboard, config: Config) -> CommandAction:
        return CommandAction.load_blackboard


class SaveReplCommand(ReplCommandBase):
    def get_name(self) -> str:
        return "save"

    def get_description(self) -> str:
        return "Save the blackboard to the local data store"

    def do(self, blackboard: Blackboard, config: Config) -> CommandAction:
        filename = input("Please enter a filename:")
        save_blackboard_to_file(blackboard=blackboard, filename=filename, config=config)

        return CommandAction.handled_already


class QuitReplCommand(ReplCommandBase):
    def get_name(self) -> str:
        return "quit"

    def get_aliases(self) -> list[str]:
        return ["bye", "exit", "stop"]

    def get_description(self) -> str:
        return "Exit the chat loop"

    def do(self, blackboard: Blackboard, config: Config) -> CommandAction:
        return CommandAction.quit


commands: list[ReplCommandBase] = [
    ClearReplCommand(),
    DumpReplCommand(),
    HelpReplCommand(),
    ListReplCommand(),
    LoadReplCommand(),
    SaveReplCommand(),
    QuitReplCommand(),
]
MIN_USER_PROMPT = 3


def _is_user_input_matching(user_prompt: str, aliases: list[str]) -> bool:
    return user_prompt.lower().strip() in aliases


def check_user_prompt(
    user_prompt: str, blackboard: Blackboard, config: Config
) -> CommandAction:
    user_prompt = user_prompt.strip()
    # special case:
    if not user_prompt or len(user_prompt) < MIN_USER_PROMPT:
        return HelpReplCommand().do(blackboard=blackboard, config=config)

    for command in commands:
        if _is_user_input_matching(
            user_prompt=user_prompt,
            aliases=command.get_aliases() + [command.get_name()],
        ):
            return command.do(blackboard=blackboard, config=config)
    return CommandAction.no_action


def print_help() -> None:
    print_assistant_message_only("Welcome to multi-agent chat")
    console.print(
        "Type in a question for the AI. If you are not sure what to type, then ask it a question like 'What can you do?'"
    )
    console.print("To exit, use the quit command")
    console.print("Available commands:")
    for command in commands:
        aliases = ", ".join(command.get_aliases())
        if aliases:
            aliases = f" (alias: {aliases})"
        console.print(
            Text(f"  {command.get_name()} - {command.get_description()}{aliases}")
        )



from dataclasses import dataclass
from enum import StrEnum, auto

from pydantic import Field

from .util_pydantic import CustomBaseModel

from .functions_dto import FunctionCallSchema
from . import util_graphql
from . import rest_api_examples


class BlackboardFormat(StrEnum):
    function_call = auto()
    graphql = auto()


class MessageRole(StrEnum):
    user = auto()
    assistant = auto()


@dataclass
class Message:
    role: MessageRole
    message: str


class FunctionCallBlackboard(CustomBaseModel):
    format: BlackboardFormat = Field(
        description="The data format of the blackboard",
        default=BlackboardFormat.function_call,
    )

    internal_previously_generated_functions: list[FunctionCallSchema] = Field(
        description="All previously generated functions: either from client (representing its data) or from agents in this generation",
        examples=[[rest_api_examples.example_create_creature_call__wolf]],
        default_factory=list,
    )
    # All previous messages in this chat (series of generations)
    internal_previous_messages: list[Message] = Field(
        default_factory=list,
        examples=[
            [
                Message(role=MessageRole.user, message="Add a wolf"),
                Message(
                    role=MessageRole.assistant,
                    message="I have created a wolf. Is there anything else I can help you with?",
                ),
            ]
        ],
    )

    # Messages that were newly-generated during this generation (required for client to know what new messages to display)
    internal_newly_generated_messages: list[Message] = Field(
        default_factory=list, examples=[[]]
    )

    # Functions that were newly-generated during this generation (required for client so they know what functions they need to execute to update their data).
    internal_newly_generated_functions: list[FunctionCallSchema] = Field(
        default_factory=list, examples=[[]]
    )

    def add_generated_functions(
        self, generated_function_calls: list[FunctionCallSchema]
    ) -> None:
        self.internal_previously_generated_functions += generated_function_calls
        self.internal_newly_generated_functions += generated_function_calls

    def get_generated_functions_matching(
        self, function_names: list[str]
    ) -> list[FunctionCallSchema]:
        return list(
            filter(
                lambda f: f.function_name in function_names,
                self.internal_previously_generated_functions,
            )
        )

    def add_message(self, message: Message) -> None:
        self.internal_previous_messages.append(message)
        self.internal_newly_generated_messages.append(message)

    def add_previous_message(self, message: Message) -> None:
        self.internal_previous_messages.append(message)

    def reset_all(self) -> None:
        """
        Completely clears out the blackboard, starting over.
        """
        self.internal_previously_generated_functions.clear()
        self.internal_previous_messages.clear()
        self.internal_newly_generated_messages.clear()
        self.internal_newly_generated_functions.clear()

    def _reset(self) -> None:
        """Resets newly created functions, to prepare for next generation"""
        self.internal_newly_generated_functions.clear()
        self.internal_newly_generated_messages.clear()

    def reset_newly_generated(self) -> None:
        """Reset newly generated, in case client did not clear out"""
        self._reset()

    def set_user_data(self, user_data: list[FunctionCallSchema]) -> None:
        """Receives the new version of user data, by setting the function-calls list, so is ready for next generation."""
        self._reset()
        self.internal_previously_generated_functions = user_data


class GraphQLBlackboard(CustomBaseModel):
    format: BlackboardFormat = Field(
        description="The data format of the blackboard",
        default=BlackboardFormat.graphql,
    )

    # Previously generated mutation calls, in this generation. Cleared out when new client data is received (so is effectively *newly* generated only).
    internal_previously_generated_mutation_calls: list[str] = Field(
        default_factory=list
    )

    # All previous messages in this chat (series of generations)
    internal_previous_messages: list[Message] = Field(default_factory=list)

    # Messages that were newly-generated during this generation (required for client to know what new messages to display)
    internal_newly_generated_messages: list[Message] = Field(default_factory=list)

    # The user data at the start of this generation.
    internal_user_data: str = Field(default="")

    def add_generated_mutations(self, generated_mutation_calls: list[str]) -> None:
        self.internal_previously_generated_mutation_calls += generated_mutation_calls

    def add_message(self, message: Message) -> None:
        self.internal_previous_messages.append(message)
        self.internal_newly_generated_messages.append(message)

    def add_previous_message(self, message: Message) -> None:
        self.internal_previous_messages.append(message)

    def get_generated_mutations_matching(
        self, accepted_graphql_schemas: list[str]
    ) -> list[str]:
        """Filter the generated mutations to suit an agent's allowed input."""
        accepted_mutation_names = util_graphql.parse_out_mutation_names_from_schemas(
            accepted_graphql_schemas
        )

        return util_graphql.filter_to_matching_mutation_calls(
            self.internal_previously_generated_mutation_calls, accepted_mutation_names
        )

    def get_user_data(self) -> str:
        return self.internal_user_data

    def set_user_data(self, user_data: str) -> None:
        self._reset()
        self.internal_user_data = user_data

    def reset_all(self) -> None:
        """
        Completely clears out the blackboard, starting over.
        """
        self.internal_previously_generated_mutation_calls.clear()
        self.internal_previous_messages.clear()
        self.internal_newly_generated_messages.clear()
        self.internal_user_data = ""

    def _reset(self) -> None:
        """
        Clears out created mutations and queries, so is ready for next generation.
        """
        self.internal_previously_generated_mutation_calls.clear()
        self.internal_newly_generated_messages.clear()

    def reset_newly_generated(self) -> None:
        """Reset newly generated, in case client did not clear out"""
        self._reset()


Blackboard = FunctionCallBlackboard | GraphQLBlackboard



import logging
from cornsnake import util_list, util_string

logger = logging.getLogger(__file__)


def _clean_mutation_name(name: str) -> str:
    result: str = util_string.filter_string_via_regex(name, "^[a-zA-Z0-9_]+$", "")
    return result


def parse_out_mutation_names_from_schemas(
    accepted_graphql_schemas: list[str],
) -> list[str]:
    accepted_mutation_names = []

    for accepted in accepted_graphql_schemas:
        token = "type Mutation {"
        if token not in accepted:
            continue
        parts = accepted.split(token)
        if len(parts) != 2:
            logger.warning("Expected 2 parts in accepted mutations schema")
            continue
        mutation_def = parts[1].split("}")[0]
        mutation = mutation_def.split("(")[0].strip()
        if mutation:
            accepted_mutation_names.append(_clean_mutation_name(mutation))

    return accepted_mutation_names


def filter_to_matching_mutation_calls(
    previously_generated_mutation_calls: list[str], accepted_mutation_names: list[str]
) -> list[str]:
    accepted_mutation_names__adjusted = util_list.flatten(
        [[f"{n}(", f"{n} ("] for n in accepted_mutation_names]
    )

    matching: list[str] = []
    for previous in previously_generated_mutation_calls:
        filtered = filter(lambda a: a in previous, accepted_mutation_names__adjusted)
        if any(list(filtered)):
            matching.append(previous)

    return matching



import logging
import time
from typing import Any, Callable
from fastapi import FastAPI, Request
from pydantic import Field

from .blackboard import FunctionCallBlackboard, Message
from .rest_api_examples import (
    FunctionAgentDefinitionMinimal,
    creature_agent_name,
    example_creeature_creator_agent,
)

from .config import load_config, Config
from .util_pydantic import CustomBaseModel

from . import prompts_router
from .agent_definition import (
    AgentDefinitionBase,
    FunctionAgentDefinition,
    build_function_agent_definition,
)

from . import main_router
from . import main_generator

logger = logging.getLogger(__file__)

app = FastAPI()


def _load_config_from_ini() -> Config:
    return load_config(path_to_ini="config.ini")


class AsyncIteratorWrapper:
    """The following is a utility class that transforms a
    regular iterable to an asynchronous one.

    link: https://www.python.org/dev/peps/pep-0492/#example-2
    """

    def __init__(self, obj: Any) -> None:
        self._it = iter(obj)

    def __aiter__(self) -> Any:
        return self

    async def __anext__(self) -> Any:
        try:
            value = next(self._it)
        except StopIteration:
            raise StopAsyncIteration
        return value


@app.middleware("http")
async def add_request_response_logging(request: Request, call_next: Callable) -> Any:
    start_time = time.perf_counter()

    config = _load_config_from_ini()
    if config.is_debug:
        print("REQUEST HEADERS", request.headers)
        request_body = await request.body()
        print("REQUEST BODY", request_body)

    response = await call_next(request)
    process_time = time.perf_counter() - start_time

    if config.is_debug:
        resp_body = [section async for section in response.__dict__["body_iterator"]]
        response.__setattr__("body_iterator", AsyncIteratorWrapper(resp_body))
        print("RESPONSE BODY", str(resp_body))

    response.headers["X-Process-Time"] = str(process_time)
    return response


class GeneratePlanRequest(CustomBaseModel):
    agent_descriptions: list[prompts_router.AgentDescription] = Field(
        description="The descriptions of the available Agents. The response will contain the most suitable agents to execute in order.",
        examples=[
            [
                {
                    "agent_name": creature_agent_name,
                    "description": "Creates new creatures given the user prompt. Ensures that ALL creatures mentioned by the user are created.",
                    "topics": ["creature"],
                    "agent_parameter_names": ["creature_name"],
                }
            ]
        ],
    )
    chat_agent_description: str = Field(
        description="Describes the 'fallback' chat agent: if no suitable agents are recommended, this chat agent will be recommended, if the user's prompt is supported. The description should include the purpose and domain of this chat system.",
        examples=["Handles users questions about an ecosystem game like Sim Life"],
    )
    user_prompt: str = Field(
        description="The input from the user",
        examples=["Add a goat instead of a sheep"],
    )
    previous_plan: prompts_router.AgentExecutionPlanSchema | None = Field(
        description="Optionally also send a previously generated plan, so the AI can generate a new plan taking into account the user's feedback (in user_prompt).",
        examples=[
            {
                "chat_message": "Certainly! I'll help you add a sheep that eats grass to your ecosystem.",
                "recommended_agents": [
                    {
                        "agent_name": creature_agent_name,
                        "rewritten_user_prompt": "Create a new creature: sheep. The sheep should have the ability to eat grass.",
                        "agent_parameters": {"creature_name": ["sheep"]},
                    }
                ],
            }
        ],
        default=None,
    )
    messages: list[Message] | None = Field(
        description="The chat message history, in case user is referring to previous messages. AI must take account of the previous messages, but prioritize the user_prompt.",
        examples=[
            [
                {"role": "user", "message": "Add grass"},
                {
                    "role": "assistant",
                    "message": "Certainly! I'll help you add grass to your ecosystem.",
                },
            ]
        ],
        default=None,
    )


class FunctionCallGenerateRequest(CustomBaseModel):
    agent_definitions: list[FunctionAgentDefinitionMinimal] = Field(
        description="The defintions of the Agents to execute, in order.",
        examples=[[example_creeature_creator_agent]],
    )
    chat_agent_description: str = Field(
        description="Describe the purpose and domain of this chat system.",
        examples=["Handles users questions about an ecosystem game like Sim Life"],
    )
    user_prompt: str = Field(
        description="The input from the user", examples=["Add a sheep that eats grass"]
    )
    blackboard: FunctionCallBlackboard | None = Field(
        description="Optionally include the previous Blackboard state, to have a conversation (avoids stateless server). This contains previous state and new data (which the user has updated either by executing its implementation of Function Calls).",
        default=None,
    )
    execution_plan: prompts_router.AgentExecutionPlanSchema | None = Field(
        description="Optionally also include a previously generated plan, to reduce latency. If no plan is included, OR there is a user prompt, then generate will also internally call generate_plan.",
        examples=[
            {
                "chat_message": "Certainly! I'll help you add a sheep that eats grass to your ecosystem.",
                "recommended_agents": [
                    {
                        "agent_name": creature_agent_name,
                        "rewritten_user_prompt": "Create a new creature: sheep. The sheep should have the ability to eat grass.",
                        "agent_parameters": {"creature_name": ["sheep"]},
                    }
                ],
            }
        ],
        default=None,
    )


@app.post("/generate_plan")
async def generate_plan(
    request: GeneratePlanRequest,
) -> prompts_router.AgentExecutionPlanSchema:
    return main_router.generate_plan_via_descriptions(
        agent_descriptions=request.agent_descriptions,
        chat_agent_description=request.chat_agent_description,
        _config=_load_config_from_ini(),
        user_prompt=request.user_prompt,
        previous_plan=request.previous_plan,
        messages=request.messages,
    )


def _build_agent_definition_from_minimal(
    minimal_agent: FunctionAgentDefinitionMinimal,
) -> FunctionAgentDefinition:
    return build_function_agent_definition(
        agent_name=minimal_agent.agent_name,
        description=minimal_agent.description,
        accepted_functions=minimal_agent.accepted_functions,
        functions_allowed_to_generate=minimal_agent.functions_allowed_to_generate,
        topics=minimal_agent.topics,
    )


@app.post("/generate_function_calls")
def generate_function_calls(
    request: FunctionCallGenerateRequest,
) -> FunctionCallBlackboard:
    agent_definitions: list[AgentDefinitionBase] = [
        _build_agent_definition_from_minimal(a) for a in request.agent_definitions
    ]

    if request.blackboard:
        request.blackboard.reset_newly_generated()  # in case client did not clear out

    blackboard = main_generator.generate_with_blackboard(
        agent_definitions=agent_definitions,
        chat_agent_description=request.chat_agent_description,
        _config=_load_config_from_ini(),
        user_prompt=request.user_prompt,
        blackboard=request.blackboard,
        execution_plan=request.execution_plan,
    )
    if not isinstance(blackboard, FunctionCallBlackboard):
        raise RuntimeError("blackboard is not FunctionCallBlackboard")
    return blackboard


# TODO (someone): Later add generate_graphql()



from atomic_agents.agents.base_agent import (
    BaseIOSchema,
)
from pydantic import Field

ParamNameToValues = dict[str, list[str]]


class GraphQLAgentInputSchema(BaseIOSchema):
    """
    This schema represents the input to the agent.
    The schema contains previously generated mutation calls, and the list of allowed generated mutations.
    """

    user_input: str = Field(description="The chat message from the user", default="")

    mutations_allowed_to_generate: list[str] = Field(
        description="Definitions of the GraphQL mutations that this agent can generate"
    )
    previously_generated_mutations: list[str] = Field(
        description="Previously generated GraphQL mutations in this chat (some are from other agents)",
        default_factory=lambda: list,
    )
    topics: list[str] = Field(
        description="This agent ONLY generates if user mentioned one of these topics"
    )
    agent_parameters: ParamNameToValues = Field(
        description="A dictionary of agent parameters there were extracted from the user's prompt.",
        default_factory=lambda: dict,
    )
    # TODO - add "If data is missing, then you need to generate allowed GraphQL queries."
    graphql_data: str = Field(
        description="The input GraphQL data from the client. You need to generate allowed GraphQL mutations that the client will use to update the data.",
        default="",
    )


class GraphQLAgentOutputSchema(BaseIOSchema):
    """
    This schema represents the output of the agent. The chat message should be non-technical - do NOT mention GraphQL.
    """

    chat_message: str = Field(
        description="The chat response to the user's message - a friendly non-technical message. Do NOT mention GraphQL."
    )
    generated_mutations: list[str] = Field(
        description="The set of new generated mutation calls to update the graphql_data. Only generate if neccessary (check if the graphql_data already has the data)."
    )
    # TODO Add generated_queries - generate queries if need more graphql_data



import instructor
from rich.console import Console
from rich.text import Text
from anthropic import AnthropicBedrock
from groq import Groq
from openai import OpenAI

from . import config

console = Console()

ClientModel = tuple[instructor.Instructor, str, int | None]  # client, model, max_tokens


def create_client(_config: config.Config) -> ClientModel:
    client: instructor.Instructor | None = None
    model: str | None = None
    max_tokens: int | None = None
    match _config.ai_platform:
        case config.AI_PLATFORM_Enum.groq:
            client = instructor.from_groq(Groq())
            model = _config.model
        case config.AI_PLATFORM_Enum.openai:
            client = instructor.from_openai(OpenAI())
            model = _config.model
        case config.AI_PLATFORM_Enum.bedrock_anthropic:
            client = instructor.from_anthropic(AnthropicBedrock())
            model = _config.model
            max_tokens = _config.max_tokens
        case _:
            raise RuntimeError(
                f"Not a recognised AI_PLATFORM: '{_config.ai_platform}' - please check Config."
            )

    console.print(Text(f"  AI platform: {_config.ai_platform}", style="magenta"))

    return client, model, max_tokens



from rich.console import Console

from . import config

console = Console()


def print_debug(message: str, config: config.Config) -> None:
    if config.is_debug:
        print(message)


def print_warning(message: str) -> None:
    console.log(f":warning: {message}", style="yellow")



import os
import logging
from cornsnake import util_dir, util_file, util_json

from pydantic import TypeAdapter
from rich.console import Console
from rich.text import Text

from .blackboard import Blackboard
from .util_pydantic import CustomBaseModel

from .config import Config

console = Console()
logger = logging.getLogger(__file__)


class SerializedBlackboard(CustomBaseModel):
    blackboard: Blackboard
    file_schema_version: str = "0.2"


def load_blackboard_from_file(
    config: Config, existing_blackboard: Blackboard
) -> Blackboard | None:
    filename = input("Please enter a filename:")
    filepath = os.path.join(config.temp_data_dir_path, filename)

    if not os.path.exists(filepath):
        console.print(
            Text(
                "That file does not exist. Please use the list command to view the current files."
            ),
            style="red",
        )
        return existing_blackboard

    console.print(f"Loading blackboard from {filepath}")
    try:
        json_data = util_json.read_from_json_file(filepath)
        serialized = TypeAdapter(SerializedBlackboard).validate_python(json_data)
        return serialized.blackboard
    except Exception as e:
        logger.exception(e)

    console.print(
        Text(
            "Could not load that blackboard (the file could be old or a different format: FunctionCalling vs GraphQL)"
        ),
        style="red",
    )
    return None


def save_blackboard_to_file(
    blackboard: Blackboard, filename: str, config: Config
) -> None:
    filename = util_file.change_extension(filename, f".{blackboard.format.value}.json")
    filepath = os.path.join(config.temp_data_dir_path, filename)

    console.print(f"Saving blackboard to {filepath}")

    serialized = SerializedBlackboard(blackboard=blackboard)
    json_data = serialized.model_dump_json()

    util_file.write_text_to_file(json_data, filepath)


def list_blackboard_files(blackboard: Blackboard, config: Config) -> None:
    files = util_dir.find_files_by_extension(
        dir_path=config.temp_data_dir_path, extension=f"{blackboard.format.value}.json"
    )
    files = [os.path.basename(f) for f in files]
    console.print(f"Blackboard data files: {files}")
